import os
import sys
import copy
import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def knn(x, k):
    inner = -2*torch.matmul(x.transpose(2, 1), x)
    xx = torch.sum(x**2, dim=1, keepdim=True)
    pairwise_distance = -xx - inner - xx.transpose(2, 1)
 
    idx = pairwise_distance.topk(k=k, dim=-1)[1]   # (batch_size, num_points, k)
    return idx


def get_input_feature(x, k=20, idx=None):
    batch_size = x.size(0)
    num_points = x.size(2)
    x = x.view(batch_size, -1, num_points)
    if idx is None:
        idx = knn(x, k=k)   # (batch_size, num_points, k)
        device = x.device

        idx_base = torch.arange(0, batch_size, device=device).view(-1, 1, 1)*num_points

        idx = idx + idx_base

        idx = idx.view(-1)
 
    _, num_dims, _ = x.size()

    x = x.transpose(2, 1).contiguous()   # (batch_size, num_points, num_dims)  -> (batch_size*num_points, num_dims) #   batch_size * num_points * k + range(0, batch_size*num_points)
    feature = x.view(batch_size*num_points, -1)[idx, :]
    feature = feature.view(batch_size, num_points, k, num_dims)  # 邻居
    x = x.view(batch_size, num_points, 1, num_dims).repeat(1, 1, k, 1)  # 自身

    #space shape
    space_shape = torch.cat((torch.sum(torch.clamp((feature-x), min=0), dim=-2, keepdim=True),
                             torch.sum(torch.clamp(-(feature-x), min=0), dim=-2, keepdim=True)), dim=-1)
    # space_shape = F.normalize(space_shape, p=2, dim=-1)  # N,6
    # space_shape = space_shape.permute(0, 2, 1).contiguous()
    # space_shape = self.space_conv(space_shape)
    # # space_shape = space_shape + space_shape_encoded
    # space_shape = space_shape.permute(0, 2, 1).contiguous()
    # space_shape = space_shape.squeeze(1)
    # space_shape = space_shape.repeat(1, 1, k, 1)
    #
    # feature = torch.cat((space_shape, x), dim=3).permute(0, 3, 1, 2).contiguous()

    return space_shape, x, idx


def get_graph_feature_p(x, k=20, idx=None):
    batch_size = x.size(0)
    num_points = x.size(2)
    x = x.view(batch_size, -1, num_points)
    if idx is None:
        idx = knn(x, k=k)  # (batch_size, num_points, k)
        device = x.device

        idx_base = torch.arange(0, batch_size, device=device).view(-1, 1, 1) * num_points

        idx = idx + idx_base

        idx = idx.view(-1)

    _, num_dims, _ = x.size()

    x = x.transpose(2,
                    1).contiguous()  # (batch_size, num_points, num_dims)  -> (batch_size*num_points, num_dims) #   batch_size * num_points * k + range(0, batch_size*num_points)
    feature = x.view(batch_size * num_points, -1)[idx, :]
    feature = feature.view(batch_size, num_points, k, num_dims)  # 邻居
    x = x.view(batch_size, num_points, 1, num_dims).repeat(1, 1, k, 1)  # 自身

    # feature = torch.cat((feature - x, x), dim=3).permute(0, 3, 1, 2).contiguous()
    d_feature = (feature - x).permute(0, 3, 1, 2).contiguous()

    return feature.permute(0, 3, 1, 2).contiguous(),d_feature, idx


def get_graph_feature_f(x, k=20, idx=None):
    batch_size = x.size(0)
    num_points = x.size(2)
    x = x.view(batch_size, -1, num_points)
    if idx is None:
        idx = knn(x, k=k)  # (batch_size, num_points, k)
        device = x.device

        idx_base = torch.arange(0, batch_size, device=device).view(-1, 1, 1) * num_points

        idx = idx + idx_base

        idx = idx.view(-1)

    _, num_dims, _ = x.size()

    x = x.transpose(2,
                    1).contiguous()  # (batch_size, num_points, num_dims)  -> (batch_size*num_points, num_dims) #   batch_size * num_points * k + range(0, batch_size*num_points)
    feature = x.view(batch_size * num_points, -1)[idx, :]
    feature = feature.view(batch_size, num_points, k, num_dims)  # 邻居
    x = x.view(batch_size, num_points, 1, num_dims).repeat(1, 1, k, 1)  # 自身

    # feature = torch.cat((feature - x, x), dim=3).permute(0, 3, 1, 2).contiguous()
    feature = feature.permute(0, 3, 1, 2).contiguous()  # 不加delta
    # feature = (feature - x).permute(0, 3, 1, 2).contiguous()

    return feature, idx


class LSPConv1(nn.Module):
    def __init__(self, in_channels, out_channels, feat_channels):
        super(LSPConv1, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.feat_channels = feat_channels

        # self.conv0 = nn.Conv2d(feat_channels, out_channels, kernel_size=1, bias=True)
        # self.conv1 = nn.Conv2d(out_channels, out_channels*in_channels, kernel_size=1, bias=True)
        # self.bn0 = nn.BatchNorm2d(out_channels)
        # self.bn1 = nn.BatchNorm2d(out_channels)
        # self.leaky_relu = nn.LeakyReLU(negative_slope=0.2)

        self.var_channels = (self.out_channels // self.in_channels) * self.in_channels
        self.conv0 = nn.Conv2d(feat_channels, self.var_channels, kernel_size=1, bias=True)
        self.conv1 = nn.Conv2d(self.var_channels, self.out_channels, kernel_size=1, bias=True)
        # self.shape_channels = (self.out_channels // 6) * 6
        self.bn0 = nn.BatchNorm2d(self.var_channels)
        self.bn1 = nn.BatchNorm2d(self.out_channels)
        self.conv_res = nn.Sequential(nn.Conv2d(self.feat_channels, self.out_channels, kernel_size=1, bias=True),
                                        nn.BatchNorm2d(self.out_channels),
                                        nn.LeakyReLU(negative_slope=0.2)
                                        )
        if self.feat_channels == 9:
            self.conv_df = nn.Sequential(nn.Conv2d(3, self.var_channels, kernel_size=1, bias=True),
                                        nn.BatchNorm2d(self.var_channels),
                                        nn.LeakyReLU(negative_slope=0.2)
                                        )
        else:
            self.conv_df = nn.Sequential(nn.Conv2d(self.feat_channels, self.var_channels, kernel_size=1, bias=True),
                                        nn.BatchNorm2d(self.var_channels),
                                        nn.LeakyReLU(negative_slope=0.2)
                                        )           

        self.leaky_relu = nn.LeakyReLU(negative_slope=0.1)
        self.sig_mlp = nn.Linear(1, 1)  # TODO 高斯的

    def forward(self, x, y, dy):
        # x: (bs, in_channels, num_points, k), y: (bs, feat_channels, num_points, k)
        batch_size, n_dims, num_points, k = x.size()

        # x repeat
        y_res = self.conv_res(y)
        x_ori = x
        x = x.repeat(1, self.out_channels // self.in_channels, 1, 1)

        x2 = torch.sum(torch.pow(x_ori, 2), 1)  # x: (bs, in_channels, num_points, k) -> (bs, num_points, k)
        dist = x2.sqrt()
        dist_max = dist.max(dim=-1, keepdim=True)[0]  # (bs, num_points, 1)
        sig = self.sig_mlp(dist_max).repeat(1,1,k)  # (bs, num_points, k)
        gaussian=torch.exp(-x2 / (2 * sig ** 2 + 1e-9))
        y = y * gaussian.unsqueeze(1).repeat(1, self.feat_channels, 1, 1)

        y = self.conv0(y) # (bs, var_channels, num_points, k)
        y = self.leaky_relu(self.bn0(y))
        # 相乘
        dy = self.conv_df(dy)
        y = y * x * dy
        y = self.conv1(y) # (bs, out_channels, num_points, k)
        y = self.bn1(y)
        y = self.leaky_relu(y)

        # y = y.permute(0, 2, 3, 1).view(batch_size, num_points, k, self.out_channels, self.in_channels) # (bs, num_points, k, out, in)
        #
        # x = x.permute(0, 2, 3, 1).unsqueeze(4) # (bs, num_points, k, in_channels, 1)
        # x = torch.matmul(y, x).squeeze(4) # (bs, num_points, k, out_channels)
        # x = x.permute(0, 3, 1, 2).contiguous() # (bs, out_channels, num_points, k)

        

        return y + y_res



class Net(nn.Module):
    def __init__(self, args, output_channels=40):
        super(Net, self).__init__()
        self.args = args
        self.k = args.k
        
        self.bn1 = nn.BatchNorm2d(64)
        self.bn2 = nn.BatchNorm2d(64)
        self.bn3 = nn.BatchNorm2d(128)
        self.bn4 = nn.BatchNorm2d(256)
        self.bn5 = nn.BatchNorm1d(args.emb_dims)

        self.conv3 = nn.Sequential(nn.Conv2d(64, 128, kernel_size=1, bias=True),
                                   self.bn3,
                                   nn.LeakyReLU(negative_slope=0.2))
        self.conv4 = nn.Sequential(nn.Conv2d(128, 256, kernel_size=1, bias=True),
                                   self.bn4,
                                   nn.LeakyReLU(negative_slope=0.2))
        self.conv5 = nn.Sequential(nn.Conv1d(512, args.emb_dims, kernel_size=1, bias=True),
                                   self.bn5,
                                   nn.LeakyReLU(negative_slope=0.2))
        self.linear1 = nn.Linear(args.emb_dims*2, 512, bias=True)
        self.bn6 = nn.BatchNorm1d(512)
        self.dp1 = nn.Dropout(p=args.dropout)
        self.linear2 = nn.Linear(512, 256)
        self.bn7 = nn.BatchNorm1d(256)
        self.dp2 = nn.Dropout(p=args.dropout)
        self.linear3 = nn.Linear(256, output_channels)

        self.lspconv1 = LSPConv1(3, 64, 9)
        self.lspconv2 = LSPConv1(3, 64, 64)

    def forward(self, x):
        batch_size = x.size(0)
        points = x

        space_shape, x, idx = get_input_feature(x, k=self.k) #BNKC
        space_shape = space_shape.permute(0, 3, 1, 2).contiguous().squeeze(-1) # BCN
        space_shape = F.normalize(space_shape, p=2, dim=1)
        space_shape = space_shape.permute(0, 2, 1).contiguous().unsqueeze(2)
        space_shape = space_shape.repeat(1, 1, self.k, 1)

        x = torch.cat((space_shape, x), dim=3).permute(0, 3, 1, 2).contiguous()
        _, dp, _ = get_graph_feature_p(points, k=self.k, idx=idx)
        x = self.lspconv1(dp, x, dp)
        x1 = x.max(dim=-1, keepdim=False)[0]

        x, dx, idx = get_graph_feature_p(x1, k=self.k)
        _, dp, _ = get_graph_feature_p(points, k=self.k, idx=idx)
        x = self.lspconv2(dp, x, dx)
        x2 = x.max(dim=-1, keepdim=False)[0]

        x, _ = get_graph_feature_f(x2, k=self.k)
        x = self.conv3(x)
        x3 = x.max(dim=-1, keepdim=False)[0]

        x, _ = get_graph_feature_f(x3, k=self.k)
        x = self.conv4(x)
        x4 = x.max(dim=-1, keepdim=False)[0]

        x = torch.cat((x1, x2, x3, x4), dim=1)

        x = self.conv5(x)
        x1 = F.adaptive_max_pool1d(x, 1).view(batch_size, -1)
        x2 = F.adaptive_avg_pool1d(x, 1).view(batch_size, -1)
        x = torch.cat((x1, x2), 1)

        x = F.leaky_relu(self.bn6(self.linear1(x)), negative_slope=0.2)
        x = self.dp1(x)
        x = F.leaky_relu(self.bn7(self.linear2(x)), negative_slope=0.2)
        x = self.dp2(x)
        x = self.linear3(x)
        return x
